#include "gnss/factor/factor_gnss_fix_2D.h"
#include "gnss/factor/factor_gnss_fix_3D.h"
#include "gnss/feature/feature_gnss_fix.h"
#include "gnss/processor/processor_gnss_fix.h"
#include "gnss/capture/capture_gnss_fix.h"

namespace wolf
{

ProcessorGnssFix::ProcessorGnssFix(ProcessorParamsGnssFixPtr _params_gnss, SensorGnssPtr _sensor_gnss_ptr) :
        ProcessorBase("GNSS FIX", _params_gnss),
        sensor_gnss_ptr_(_sensor_gnss_ptr),
        params_gnss_(_params_gnss),
        first_capture_ptr_(nullptr)
{
    //
}

ProcessorGnssFix::~ProcessorGnssFix()
{
    //
}

void ProcessorGnssFix::processCapture(CaptureBasePtr _capture_ptr)
{
    // TODO: keep captures in a buffer and deal with KFpacks

    //WOLF_DEBUG("ProcessorGnssFix::process()");
    last_capture_ptr_ = std::static_pointer_cast<CaptureGnssFix>(_capture_ptr);

    // DEBUGGING ENU_ECEF
    //Eigen::Vector3s p = last_capture_ptr_->getData();
    //Eigen::Vector3s p_enu = sensor_gnss_ptr_->getREnuEcef()*p + sensor_gnss_ptr_->gettEnuEcef();
    //std::cout << "WOLF: gnss fix in ENU coordinates: " << p_enu.transpose() << std::endl;

    FrameBasePtr new_frame_ptr = nullptr;
    FactorBasePtr new_fac = nullptr;

    // ALREADY CREATED KF
    PackKeyFramePtr KF_pack = buffer_pack_kf_.selectPack( _capture_ptr, params_gnss_->time_tolerance);
    if (KF_pack && KF_pack->key_frame != last_gnss_fix_KF_)
    {
        WOLF_DEBUG( "PR ", getName()," - capture ", _capture_ptr->id(), " appended to existing KF " , KF_pack->key_frame->id() , " TS: ", KF_pack->key_frame->getTimeStamp());
        new_frame_ptr = KF_pack->key_frame;
    }
    // MAKE KF
    else if (voteForKeyFrame() && permittedKeyFrame())
    {
        new_frame_ptr = getProblem()->emplaceFrame(KEY, _capture_ptr->getTimeStamp());
        getProblem()->keyFrameCallback(new_frame_ptr, shared_from_this(), params_gnss_->time_tolerance);
        WOLF_DEBUG( "PR ", getName()," - capture ", _capture_ptr->id(), " appended to new KF " , new_frame_ptr->id() , " TS: ", new_frame_ptr->getTimeStamp());
    }

    // ESTABLISH FACTOR
    if (new_frame_ptr)
    {
        // LINK CAPTURE
        _capture_ptr->link(new_frame_ptr); // Add incoming Capture to the new Frame

        // EMPLACE FEATURES
        //WOLF_DEBUG( "PR ", getName()," - emplacing the feature...");
        auto ftr_ptr = FeatureBase::emplace<FeatureGnssFix>(last_capture_ptr_, last_capture_ptr_->getData(),last_capture_ptr_->getDataCovariance());

        // EMPLACE FACTOR
        new_fac = emplaceFactor(ftr_ptr);

        // outlier rejection
        if (sensor_gnss_ptr_->isEnuDefined() && sensor_gnss_ptr_->isEnuMapInitialized())
            if (rejectOutlier(new_fac))
                new_fac = nullptr;

        // store last KF
        if (new_fac)
            last_gnss_fix_KF_= new_frame_ptr;
    }

    // INITIALIZE ENU IF:
    //      1 - not initialized
    //      2 - factor established
    if (!sensor_gnss_ptr_->isEnuDefined() && new_fac != nullptr)
    {
        sensor_gnss_ptr_->setEcefEnu(last_capture_ptr_->getData());

        // Store the first capture that established a factor
        first_capture_ptr_ = last_capture_ptr_;
    }

    // INITIALIZE ENU_MAP IF 4 NECESSARY CONDITIONS:
    //      1 - ENU-ECEF defined
    //      2 - not initialized
    //      3 - current capture in key-frame with factor
    //      4 - two factors established (first and current) in frames separated enough ( > enu_map_init_dist_min)
    if ( sensor_gnss_ptr_->isEnuDefined() &&
        !sensor_gnss_ptr_->isEnuMapInitialized() &&
         new_fac != nullptr &&
         first_capture_ptr_->getFrame() != nullptr && last_capture_ptr_->getFrame() != nullptr &&
         (first_capture_ptr_->getFrame()->getState()-last_capture_ptr_->getFrame()->getState()).norm() > params_gnss_->enu_map_init_dist_min)
    {
        std::cout << "initializing enu map\n";
        sensor_gnss_ptr_->initializeEnuMap(first_capture_ptr_->getFrame()->getState(), first_capture_ptr_->getData(),
                                           last_capture_ptr_->getFrame()->getState(), last_capture_ptr_->getData());
    }
}

FactorBasePtr ProcessorGnssFix::emplaceFactor(FeatureBasePtr& ftr_ptr)
{
    // CREATE CONSTRAINT --------------------
    //WOLF_DEBUG("creating the factor...");
    // 2D
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssFix2D>(ftr_ptr, ftr_ptr, sensor_gnss_ptr_, shared_from_this(), false, FAC_ACTIVE);
    // 3D
    else
        return FactorBase::emplace<FactorGnssFix3D>(ftr_ptr, ftr_ptr, sensor_gnss_ptr_, shared_from_this(), false, FAC_ACTIVE);
}

bool ProcessorGnssFix::rejectOutlier(FactorBasePtr fac_ptr)
{
    // Cast feature
    auto gnss_ftr_ptr = std::static_pointer_cast<FeatureGnssFix>(fac_ptr->getFeature());
    // copy states
    Eigen::VectorXs x(gnss_ftr_ptr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXs o(gnss_ftr_ptr->getCapture()->getFrame()->getP()->getState());
    Eigen::VectorXs x_antena(sensor_gnss_ptr_->getStateBlock(0)->getState());
    Eigen::VectorXs t_ENU_map(sensor_gnss_ptr_->getEnuMapTranslation()->getState());
    Eigen::VectorXs roll_ENU_map(sensor_gnss_ptr_->getEnuMapRoll()->getState());
    Eigen::VectorXs pitch_ENU_map(sensor_gnss_ptr_->getEnuMapPitch()->getState());
    Eigen::VectorXs yaw_ENU_map(sensor_gnss_ptr_->getEnuMapYaw()->getState());
    // create Scalar* array of a copy of the state
    Scalar* parameters[7] = {x.data(), o.data(), x_antena.data(), t_ENU_map.data(), roll_ENU_map.data(),
                             pitch_ENU_map.data(), yaw_ENU_map.data()};
    // create residuals pointer
    Eigen::VectorXs residuals(3);
    // evaluate the factor in this state
    fac_ptr->evaluate(parameters, residuals.data(), nullptr);
    // discard if residual too high evaluated at the current estimation
    if (residuals.norm() > 1e3)
    {
        WOLF_WARN("Discarding GNSS FIX Factor, considered OUTLIER");
        WOLF_TRACE("Feature: ", fac_ptr->getMeasurement().transpose(),"\nError: ",(fac_ptr->getMeasurementSquareRootInformationUpper().inverse()*residuals).transpose());
        fac_ptr->remove();
        return true;
    }
    return false;
}

bool ProcessorGnssFix::voteForKeyFrame()
{
    // Depending on time since the last KF with gnssfix capture
    if (!last_gnss_fix_KF_ || (last_capture_ptr_->getTimeStamp() - last_gnss_fix_KF_->getTimeStamp()) > params_gnss_->time_th)
        return true;

    // TODO: more alternatives?

    // otherwise
    return false;
}

void ProcessorGnssFix::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ptr_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

ProcessorBasePtr ProcessorGnssFix::create(const std::string& _unique_name, const ProcessorParamsBasePtr _params, const SensorBasePtr sensor_ptr)
{
    auto prc_ptr = std::make_shared<ProcessorGnssFix>(std::static_pointer_cast<ProcessorParamsGnssFix>(_params),
                                                      std::static_pointer_cast<SensorGnss>(sensor_ptr));
    prc_ptr->setName(_unique_name);
    return prc_ptr;
}


} // namespace wolf


// Register in the SensorFactory
#include "core/processor/processor_factory.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR("GNSS FIX",ProcessorGnssFix)
} // namespace wolf
