#include "gnss/factor/factor_gnss_single_diff_2D.h"
#include "gnss/feature/feature_gnss_single_diff.h"
#include "gnss/processor/processor_gnss_single_diff.h"
#include "gnss/capture/capture_gnss_single_diff.h"

namespace wolf
{

ProcessorGnssSingleDiff::ProcessorGnssSingleDiff(ProcessorParamsGnssSingleDiffPtr _params_gnss, SensorGnssPtr _sensor_gnss_ptr) :
        ProcessorBase("GNSS SINGLE DIFFERENCES", _params_gnss),
        sensor_gnss_ptr_(_sensor_gnss_ptr),
        params_gnss_(_params_gnss)
{
    //
}

ProcessorGnssSingleDiff::~ProcessorGnssSingleDiff()
{
    //
}

void ProcessorGnssSingleDiff::processCapture(CaptureBasePtr _capture_ptr)
{
    // TODO: keep captures in a buffer and deal with KFpacks

    //WOLF_DEBUG("ProcessorGnssSingleDiff::process()");
    incoming_capture_ptr_ = std::static_pointer_cast<CaptureGnssSingleDiff>(_capture_ptr);

    // discard capture with null or non-key origin frame
    if (incoming_capture_ptr_->getOriginFrame() == nullptr || incoming_capture_ptr_->getOriginFrame()->isKey())
        return;

    FrameBasePtr new_frame_ptr = nullptr;

    // ALREADY CREATED KF
    PackKeyFramePtr KF_pack = buffer_pack_kf_.selectPack( _capture_ptr, params_->time_tolerance);
    if (KF_pack && KF_pack->key_frame != incoming_capture_ptr_->getOriginFrame())
    {
        new_frame_ptr = KF_pack->key_frame;
        WOLF_DEBUG( "PR ",getName()," - capture ", _capture_ptr->id(), " appended to existing KF " , KF_pack->key_frame->id() , " TS: ", KF_pack->key_frame->getTimeStamp());
    }
    // MAKE KF
    else if (voteForKeyFrame() && permittedKeyFrame())
    {
        new_frame_ptr = getProblem()->emplaceFrame(KEY, _capture_ptr->getTimeStamp());
        getProblem()->keyFrameCallback(new_frame_ptr, shared_from_this(), params_->time_tolerance);
        WOLF_DEBUG( "PR ",getName()," - capture ", _capture_ptr->id(), " appended to new KF " , new_frame_ptr->id() , " TS: ", new_frame_ptr->getTimeStamp());
    }

    // ESTABLISH FACTOR
    if (new_frame_ptr)
    {
        // LINK CAPTURE
        _capture_ptr->link(new_frame_ptr); // Add incoming Capture to the new Frame

        // EMPLACE FEATURE
        //WOLF_DEBUG("emplacing feature...");
        auto ftr_ptr = FeatureBase::emplace<FeatureGnssSingleDiff>(incoming_capture_ptr_, incoming_capture_ptr_->getData(),incoming_capture_ptr_->getDataCovariance());

        // ADD CONSTRAINT
        //WOLF_DEBUG("emplacing factor...");
        emplaceFactor(ftr_ptr);

        //WOLF_DEBUG("Factor established");
    }

    // INITIALIZE ENU_MAP IF 4 NECESSARY CONDITIONS:
    //      1 - ENU-ECEF defined
    //      2 - not initialized
    //      3 - current capture in key-frame with factor
    //      4 - frames constained by the factor separated enough ( > enu_map_init_dist_min)
    if ( sensor_gnss_ptr_->isEnuDefined() &&
        !sensor_gnss_ptr_->isEnuMapInitialized() &&
         new_frame_ptr != nullptr &&
         incoming_capture_ptr_->getFrame() != nullptr && incoming_capture_ptr_->getFrame()->isKey() &&
         incoming_capture_ptr_->getData().norm() > params_gnss_->enu_map_init_dist_min)
    {
        std::cout << "initializing enu map\n";
        sensor_gnss_ptr_->initializeEnuMapYaw(last_capture_ptr_->getOriginFrame()->getState(),
                                              last_capture_ptr_->getFrame()->getState(),
                                              last_capture_ptr_->getData());
    }

    last_capture_ptr_ = incoming_capture_ptr_;
    incoming_capture_ptr_ = nullptr;
}

FactorBasePtr ProcessorGnssSingleDiff::emplaceFactor(FeatureBasePtr& ftr_ptr)
{
    // CREATE CONSTRAINT --------------------
    //WOLF_DEBUG("creating the factor...");
    // 2D
    if (getProblem()->getDim() == 2)
        return FactorBase::emplace<FactorGnssSingleDiff2D>(ftr_ptr, ftr_ptr, incoming_capture_ptr_->getOriginFrame(), sensor_gnss_ptr_, shared_from_this());
    // 3D TODO
    else
        std::runtime_error("Single Differences in 3D not implemented yet.");

    return nullptr;
}

bool ProcessorGnssSingleDiff::voteForKeyFrame()
{
    // Depending on time since the last KF with gnssfix capture
    if (last_capture_ptr_==nullptr || (last_capture_ptr_->getTimeStamp() - incoming_capture_ptr_->getTimeStamp()) > params_gnss_->time_th)
        return true;

    // TODO: more alternatives?

    // otherwise
    return false;
}

void ProcessorGnssSingleDiff::configure(SensorBasePtr _sensor)
{
    sensor_gnss_ptr_ = std::static_pointer_cast<SensorGnss>(_sensor);
};

ProcessorBasePtr ProcessorGnssSingleDiff::create(const std::string& _unique_name, const ProcessorParamsBasePtr _params, const SensorBasePtr sensor_ptr)
{
    ProcessorGnssSingleDiffPtr prc_ptr = std::make_shared<ProcessorGnssSingleDiff>(std::static_pointer_cast<ProcessorParamsGnssSingleDiff>(_params), std::static_pointer_cast<SensorGnss>(sensor_ptr));
    prc_ptr->setName(_unique_name);
    return prc_ptr;
}

} // namespace wolf


// Register in the SensorFactory
#include "core/processor/processor_factory.h"
namespace wolf {
WOLF_REGISTER_PROCESSOR("GNSS SINGLE DIFFERENCES",ProcessorGnssSingleDiff)
} // namespace wolf
